var express = require('express');
var path = require('path');
var moment = require('moment');
var AsnHelper = require("../components/helpers/AsnHelper");

var json2csv = require('json2csv');
//var exportExcel = require("../components/helpers/exportODS");
var mssql = require('mssql');
var env = process.env.NODE_ENV || "prod";
var configJson = require(path.join(__dirname, '..', '/components/config/config.json'))[env];
var _ = require("lodash");

var config = {
    server: configJson.host,
    user: configJson.username,
    password: configJson.password,
    database: configJson.database,
    stream: true,
    port: configJson.port,
    requestTimeout: configJson.dialectOptions.requestTimeout,
    options: {
        encrypt: true
    }
};

mssql.on('error', function(err) {
    logger.error("sql error: " + err);
});


module.exports = function(app, passport, models) {

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');

    var env         = process.env.NODE_ENV || "dev";
    var config      = require("../components/config/config.json")[env];

    /**
     * route to create a web-only user
     */
    app.post('/api/signup', passport.authenticate('signup', {
        successRedirect: '/index',
        failureRedirect: '/'
    }));

    /**5:44
     * route to test if the user is logged in or not
     */
    app.get('/api/loggedin/', function(req, res) {
        res.send(req.isAuthenticated() ? req.user : '0');
    });

    /**
     * route to log in
     */
    app.post('/api/login', function(req, res, next) {
        passport.authenticate('local', function(err, user) {
            if (err) {
                logger.error("login error: " + err);
                return next(err);}
            if (!user) {
                logger.error("no user found");

                return next("No User Found. Please Try Again.");
            }
            logger.debug("logging in...");
            //now use passport to actually log in
            req.login(user, function(err) {

                    if (err) {
                        logger.error("req.login error: " + err);

                        return next(err);
                    }
                    logger.debug("req.login user: " + req.user);

                    res.send(req.user);
                }
            );
        })(req, res, next);
    });

    /**
     * route to log out
     */
    app.post('/api/logout', function(req, res){
        req.logOut();
        res.send(200);
    });

    /**
     * Enforce authentication on all routes.
     * If the main index page is requested, send back the countData page
     */
    app.get('*', auth, function (req, res, next) {
        res.sendFile('components/public/countData/countData.html', {root: path.join(__dirname, '..')});
    });

    /**************    Helpers for Web Application**********************/

    //Used to log browser errors in the server log
    app.post('/api/saveLogData', function(req, res) {
        var msg = "";
        try {
            msg = JSON.stringify(req.body);
        } catch (e) {
            msg = "stringify error";
        }
        logger.debug("log from api: " + msg);
        res.status(200).send("");
    });



    /**
     * Common routine for external filtering the web grids from MS SQL database
     */
    function buildWhereArgsMsSql(key, newCondition, objName, objRawAttributes) {
        var whereArg = null;
        var dataType = String(objRawAttributes[key].type);
        if (dataType.indexOf("VARCHAR") < 0) {
            //cast numerics as strings so we can use the "LIKE" operator
            if (dataType.indexOf("TIME") > 0) {
                //since time is hidden in the UI, we should strip time from the date fields
                if (key === 'expiration') {
                    whereArg = models.Sequelize.where(models.Sequelize.fn('FORMAT', models.Sequelize.cast(models.Sequelize.col(objName + "." + objRawAttributes[key].field), 'DATE'), 'MM / yy', 'en-US'), 'LIKE', newCondition);
                } else {
                    whereArg = models.Sequelize.where(models.Sequelize.cast(models.Sequelize.cast(models.Sequelize.col(objName + "." + objRawAttributes[key].field), 'DATE'), 'varchar'), 'LIKE', newCondition);
                }
            } else {
                whereArg = models.Sequelize.where(models.Sequelize.cast(models.Sequelize.col(objName + "." + objRawAttributes[key].field), 'varchar'), 'LIKE', newCondition);
            }
        } else {
            //already a String... just use the LIKE operator
            whereArg = {};
            whereArg[objRawAttributes[key].field] = {$like: newCondition};
        }

        return whereArg;
    }

    /**
     * Common routine for external filtering the web grids from MySQL database
     */
    function buildWhereArgsMySql(key, newCondition, objName, objRawAttributes) {
        var whereArg = null;
        var dataType = String(objRawAttributes[key].type);
        if (dataType.indexOf("VARCHAR") < 0) {
            //cast numerics as strings so we can use the "LIKE" operator
            if (dataType.indexOf("TIME") > 0) {
                //since time is hidden in the UI, we should strip time from the date fields
                whereArg = models.Sequelize.where(models.Sequelize.fn('DATE_FORMAT', models.Sequelize.col(objName + "." + objRawAttributes[key].field) ,'%m-%d-%Y'), 'LIKE', newCondition);
            } else {
                whereArg = models.Sequelize.where(models.Sequelize.cast(models.Sequelize.col(objName + "." + objRawAttributes[key].field), 'CHAR(255)'), 'LIKE', newCondition);
            }
        } else {
            //already a String... just use the LIKE operator
            whereArg = {};
            whereArg[objRawAttributes[key].field] = {$like: newCondition};
        }

        return whereArg;
    }

    function getWhereArgForDb(key, newCondition, objName, objRawAttributes) {
        if (configJson.dialect === 'mssql') {
            return buildWhereArgsMsSql(key, newCondition, objName, objRawAttributes);
        }
        return buildWhereArgsMySql(key, newCondition, objName, objRawAttributes);
    }

    function buildWhereArgs(req, objName, objRawAttributes) {
        var whereArgs = [];
        if (req.body.where != null) {
            console.log(JSON.stringify(req.body.where));
            for (var key in req.body.where) {
                if (objRawAttributes.hasOwnProperty(key)) {
                    var newCondition = String(req.body.where[key]);
                    if (newCondition === "[object Object]") {
                        if (key === "timestamp") {
                            //we have gt/lt range
                            if (req.body.where[key]['$gt']) {
                                var newWhere = {};
                                newWhere[objRawAttributes[key].field] = {$gt: moment(req.body.where[key]['$gt'])};
                                whereArgs.push(newWhere);
                            }
                            if (req.body.where[key]['$lt']) {
                                var newWhere = {};
                                newWhere[objRawAttributes[key].field] = {$lt: moment(req.body.where[key]['$lt'])};
                                whereArgs.push(newWhere);
                            }
                            continue;
                        } else {
                            newCondition = String(req.body.where[key]['$like']);
                        }
                    }
                    var dataType = String(objRawAttributes[key].type);
                    if (newCondition.indexOf("*") < 0) {
                        newCondition = "%" + newCondition + "%";
                    } else {
                        newCondition = newCondition.replace(/\*/g, "%");
                    }
                    whereArgs.push(getWhereArgForDb(key, newCondition, objName, objRawAttributes));
                    if (dataType.indexOf("VARCHAR") < 0) {
                        delete req.body.where[key];
                    }
                }
            };
        }
        return whereArgs;
    }
    //CountData Grid
    function buildCountDataWhereArgs(req) {
        var args = buildWhereArgs(req, 'CountData', models.CountData.rawAttributes);
        var newWhere = {};
        newWhere[models.CountData.rawAttributes['deviceId'].field] = {$ne : null };
        args.push(newWhere);
        return args;
    }
    //Inventory/PO/Receipt/Consumption Grids
    function buildInvWhereArgs(req) {
        return buildWhereArgs(req, 'Inventory', models.Inventory.rawAttributes);
    }
    //FTP Grid
    function buildFTPWhereArgs(req) {
        return buildWhereArgs(req, 'FTPLog', models.FTPLog.rawAttributes);
    }
    //Login Messages Grid
    function buildMessageWhereArgs(req) {
        return buildWhereArgs(req, 'Message', models.Message.rawAttributes);
    }
    //NDC (POC) Grid
    function buildNdcWhereArgs(req) {
        return buildWhereArgs(req, 'NdcData', models.NdcData.rawAttributes);
    }


    /************************* APIs for Web Application *************************/

    /**
     * Return a total count of Count Data items that meet the filter criteria
     * (for pagination)
     *
     * Params:  (Body) "where" object with attributes that match the CountData model
     */
    app.post('/api/getCountDataTotal/', function(req, res) {
        var whereArgs = buildCountDataWhereArgs(req);

        models.CountData.count({
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset
            }
        ).then(function(count) {
            res.send(String(count));
        })
    });

    /**
     * Return the actual data that meets the filter criteria
     *
     * Params:  (Body) "where" object with attributes that match the CountData model
     */
    app.post('/api/getCountData/', function(req, res) {
        var whereArgs = buildCountDataWhereArgs(req);


        models.CountData.findAll(
            {
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset,
                order: req.body.order
            }

        ).then(function (countData) {
            res.send(countData);
        });
    });

    /**
     * Return a total count of On Hand Inventory Data items that meet the where criteria
     * (for pagination)
     *
     * Params:  (Body) "where" object with attributes that match the Inventory model
     */
    app.post('/api/getInventoryTotal/', function(req, res) {
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'INV'});

        models.Inventory.count({
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset
            }
        ).then(function(count) {
            res.send(String(count));
        })
    });

    /**
     * Return all On Hand Inventory data that meets the filter criteria
     *
     * Params:  (Body) "where" object with attributes that match the Inventory model
     */
    app.post('/api/getInventory/', function(req, res) {
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'INV'});

        models.Inventory.findAll(
            {
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset,
                order: req.body.order
            }

        ).then(function (inventory) {
            res.send(inventory);
        });
    });

    /**
     * Return a total count of PO Data items that meet the where criteria
     * (for pagination)
     *
     * Params:  (Body) "where" object with attributes that match the Inventory model
     */
    app.post('/api/getPoTotal/', function(req, res) {
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'PO'});

        models.Inventory.count({
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset
            }
        ).then(function(count) {
            res.send(String(count));
        })
    });

    /**
     * Return all PO data that meets the filter criteria
     *
     * Params:  (Body) "where" object with attributes that match the Inventory model
     */
    app.post('/api/getPo/', function(req, res) {
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'PO'});
        models.Inventory.findAll(
            {
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset,
                order: req.body.order
            }

        ).then(function (inventory) {
            res.send(inventory);
        });
    });

    /**
     * Return a total count of Receipts Data items that meet the where criteria
     * (for pagination)
     *
     * Params:  (Body) "where" object with attributes that match the Inventory model
     */
    app.post('/api/getReceiptTotal/', function(req, res) {
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'RECEIPT'});

        models.Inventory.count({
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset
            }
        ).then(function(count) {
            res.send(String(count));
        })
    });

    /**
     * Return all Receipt data that meets the filter criteria
     *
     * Params:  (Body) "where" object with attributes that match the Inventory model
     */
    app.post('/api/getReceipt/', function(req, res) {
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'RECEIPT'});
        var out = {
            count: 0,
            rows: null
        };
        models.Inventory.count({
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset
            }
        ).then(function(count) {

            out.count = count;

            models.Inventory.findAll(
                {
                    where: whereArgs,
                    limit: req.body.limit,
                    offset: req.body.offset,
                    order: req.body.order
                }

            ).then(function (receipt) {
                out.data = receipt;
                res.header("X-Total-Count", out.count);
                res.status(200).send(out.data);
            }).catch(function(ferr){
                res.status(500).send(JSON.stringify(ferr));
            });
        }).catch(function(cerr){
            res.status(500).send(JSON.stringify(cerr));
        });
    });

    /**
     * Saves just the receipt qty
     */
    app.post('/api/v2/saveReceiptQty', function(req, res){
        models.Inventory.findOne({ where: { id: req.body.id }}).then(function(receipt){
            if (!receipt) {
                res.status(500).send("Receipt not found: ");
            } else {
                receipt.count = req.body.count;
                receipt.modifiedBy = req.body.modifiedBy;
                receipt.modifiedDate = req.body.modifiedDate;
                transactionLogger.debug("saving inventory record: " + JSON.stringify(receipt));
                receipt.save().then(function(newReceipt){
                    res.status(200).send(newReceipt);
                }).catch(function(serr){
                    res.status(500).send("Unable to save Receipt: " + JSON.stringify(serr));
                })
            }
        }).catch(function(err){
            res.status(500).send("Receipt not found: " + JSON.stringify(err));
        })
    });

    /**
     * Return a total count of Consumption Data items that meet the where criteria
     * (for pagination)
     *
     * Params:  (Body) "where" object with attributes that match the Inventory model
     */
    app.post('/api/getConsumptionTotal/', function(req, res) {
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'CONSUMERUSAGE'});

        models.Inventory.count({
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset
            }
        ).then(function(count) {
            res.send(String(count));
        })
    });


    /**
     * Return all Consumption data that meets the filter criteria
     *
     * Params:  (Body) "where" object with attributes that match the Inventory model
     */
    app.post('/api/getConsumption/', function(req, res) {
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'CONSUMERUSAGE'});
        models.Inventory.findAll(
            {
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset,
                order: req.body.order
            }

        ).then(function (consumption) {
            res.send(consumption);
        });
    });

    /**
     * Used for UI or REST console to list FTP log entries
     *
     * Params:  (Body) "where" object with attributes that match the FTP model
     */
    app.post('/api/getFTPdata', function(req, res) {
        var whereArgs = buildFTPWhereArgs(req);
        var ob = req.body.order;
        if (!ob) {
            ob = '"sent" DESC';
        }
        models.FTPLog.findAll(
            {
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset,
                order: ob
            }
        ).then(function(logs) {
            res.send(logs);
        }).catch(function(err){
            console.log(err);
            res.sendStatus(500);
        });
    });

    /**
     * Return a total count of FTP Data items that meet the where criteria
     * (for pagination)
     *
     * Params:  (Body) "where" object with attributes that match the FTP model
     */
    app.post('/api/getFTPDataTotal', function(req, res) {
        var whereArgs = buildFTPWhereArgs(req);
        var ob = req.body.order;
        if (!ob) {
            ob = '"sent" DESC';
        }
        models.FTPLog.count(
            {
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset,
                order: ob
            }
        ).then(function(logs) {
            res.send(logs);
        }).catch(function(err){
            console.log(err);
            res.sendStatus(500);
        });
    });

    /**
     * Returns all Vendors in the table
     */
    app.post('/api/getVendorList', function(req, res) {
        models.VendorList.findAll({order: '"vendorDisplay" ASC'}).then(function(vendors) {
            res.send(vendors);
        });
    });

    /**
     * Return All OR a specific Vendor
     *
     * Optional Params:  (Body) vendorId
     */
    app.post('/api/getVendor', function(req, res) {
        if (req.body.vendorId == null) {
            models.VendorList.findAll()
                .then(function (items) {
                    res.send(items);
                });
        } else {
            models.VendorList.findAll({
                where: {
                    $or: [
                        {vendorId: req.body.vendorId},
                        {vendorDisplay: req.body.vendorId}
                    ]
                }
            }).then(function (items) {
                res.send(items);
            });
        }
    });

    /**
     * Returns all Catalogs in the table
     */
    app.post('/api/getCatalogList', function(req, res) {
        models.CatalogList.findAll({order: '"catalogDisplay" ASC'}).then(function(vendors) {
            res.send(vendors);
        });
    });

    /**
     * Return All OR a specific Catalog
     *
     * Optional Params:  (Body) catalogId
     */
    app.post('/api/getCatalog', function(req, res) {
        if (req.body.catalogId == null) {
            models.CatalogList.findAll()
                .then(function (items) {
                    res.send(items);
                });
        } else {
            models.CatalogList.findAll({
                where: {
                    $or: [
                        {catalogId: req.body.catalogId},
                        {catalogDisplay: req.body.catalogId}
                    ]
                }
            }).then(function (items) {
                res.send(items);
            });
        }
    });

    /**
     * Return a total count of Login message Data items that meet the filter criteria
     * (for pagination)
     *
     * Params:  (Body) "where" object with attributes that match the Message model
     */
    app.post('/api/getMessagesTotal/', function(req, res) {
        var whereArgs = buildMessageWhereArgs(req);

        models.Message.count({
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset
            }
        ).then(function(count) {
            res.send(String(count));
        })
    });

    /**
     * Return all messages that meet the filter criteria
     *
     * Params:  (Body) "where" object with attributes that match the Message model
     */
    app.post('/api/getMessages/', function(req, res) {
        var whereArgs = buildMessageWhereArgs(req);

        models.Message.findAll(
            {
                where: whereArgs,
                include: [
                    {model: models.User, as: "createdBy", attributes: ['displayName']},
                    {model: models.MsgRecipient, as: "recipients", include: [ {model: models.ClinicProfile, as: "clinic", attributes:['clinicId']}]}
                    ],
                limit: req.body.limit,
                offset: req.body.offset,
                order: req.body.order
            }

        ).then(function (mData) {
            res.send(mData);
        });
    });

    //create a message
    app.post('/api/message', function(req, res) {

        var inJson = req.body;
        var outJson = {};
        if (inJson.id) {
            res.status("400").send("Invalid method")
        }
        else if (!inJson.recipients || !Array.isArray(inJson.recipients)) {
            res.status("400").send("At least one recipient is required")
        }
        else {
            var msg = models.Message.build({
                message: inJson.message,
                timestamp: new Date(),
                createdByUserId: inJson.createdByUserId,
                keepAliveStart: inJson.keepAliveStart,
                keepAliveEnd: inJson.keepAliveEnd
            });

            msg.save().then(function(savedMsg){
                outJson = savedMsg;
                var promises = [];
                var id = savedMsg.id;
                for (var r =0; r < inJson.recipients.length; r++) {
                    var recipient = inJson.recipients[r];
                    var recip = models.MsgRecipient.build({
                        msgId: id,
                        clinicId: recipient.clinicId
                    });
                    promises.push(recip.save())
                }
                models.sequelize.Promise.all(promises).then(function(children){
                    outJson.recipients = children;
                    res.send(outJson);
                }).catch(function(recipErr){
                    res.status(500).send(recipErr);
                });
            }).catch(function(msgErr){
                res.status(500).send(msgErr);
            });
        }

    });

    app.put('/api/message/:id', function(req, res) {
        if (req.params && req.params['id']){
            var inJson = req.body;
            var outJson = {};
            if (!inJson.recipients || !Array.isArray(inJson.recipients)) {
                res.status("400").send("At least one recipient is required")
            }
            else {
                models.Message.findOne({
                    where: { id: req.params['id'] }
                }).then(function(existingMessage){
                    existingMessage.message = inJson.message;
                    existingMessage.createdByUserId = inJson.createdByUserId;
                    existingMessage.keepAliveStart = inJson.keepAliveStart;
                    existingMessage.keepAliveEnd = inJson.keepAliveEnd;
                    existingMessage.save().then(function(newMessage){
                        outJson = newMessage;
                        models.MsgRecipient.destroy({
                            where: { msgId: newMessage.id }
                        }).then(function(count){
                            var promises = [];
                            for (var r =0; r < inJson.recipients.length; r++) {
                                var recipient = inJson.recipients[r];
                                var recip = models.MsgRecipient.build({
                                    msgId: id,
                                    clinicId: recipient.clinicId
                                });
                                promises.push(recip.save())
                            }
                            models.sequelize.Promise.all(promises).then(function(children){
                                outJson.recipients = children;
                                res.send(outJson);
                            }).catch(function(recipErr){
                                res.status(500).send(recipErr);
                            });
                        }).catch(function(delErr){
                            res.status(500).send(delErr);
                        });
                    }).catch(function(saveErr){
                        res.status(500).send(saveErr);
                    });
                }).catch(function(msgErr){
                    res.status(500).send(msgErr);
                });
            }
        } else {
            res.status(400).send("Invalid id");
        }
    });

    //Special API for add or update functionality
    app.post('/api/v2/saveMessage', function(req, res) {
        console.log("request: " + req.session);
        upsertMessage(req.body, req)
            .then(function(result) {
                if (!result) result = req.body;
                res.send(result);
            });
    });

    app.post('/api/v2/deleteMessage', function(req, res) {
        models.Message.destroy({
            where: { id: req.body.id }
        }).then(function (){
            transactionLogger.info("Message deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
            res.send(req.body);
        });
    });



    /**
     *  Update or Insert a record. Searches for a record, if found, updates it. If not found, creates it then updates it.
     *  Returns 1 if updated, 0 if created, -1 for error (but not really)
     */
    function upsertMessage(message, req) {
        if (message.hasOwnProperty("recipients")) delete message.recipients;
        return models.Message.findOrCreate({where: {id: message.id}, defaults: {message: message.message, createdByUserId: message.createdByUserId}, raw: true })
            .spread(function(itemRecord, created) {
                message.id = itemRecord.id;
                models.Message.update(message, {where: {id: message.id}})
                    .then(function (count) {
                        var recCount = count;
                        if (Array.isArray(count)) recCount = count[0];
                        if (recCount >= 0 && message.Recipients) {
                            var ids = [];
                            for (var r=0; r < message.Recipients.length;r++){
                                ids.push(message.Recipients[r].id);
                            }
                            models.MsgRecipient.destroy({where: {messageId: message.id}}).then(function (num) {
                                var cats = [];
                                for (var i = 0; i < ids.length; i++) {
                                    var cat = models.MsgRecipient.build({
                                        messageId: message.id,
                                        clinicId: ids[i]
                                    });
                                    cats.push(cat);
                                }
                                if (cats.length > 0) {
                                    var promises = [];

                                    cats.forEach(function (cat) {
                                        promises.push(cat.save());
                                    });

                                    models.sequelize.Promise.all(promises).then(function (vals) {
                                        transactionLogger.info("Message " + ((created) ? "created" : "updated") + " by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(message));
                                        return cats.length;
                                    }).catch(function (saveErr) {
                                        logger.error(saveErr);
                                        return -1;
                                    });
                                }
                            }).catch(function (err) {
                                logger.error(err);
                                return -1;
                            });
                        }
                    }).catch(function (err) {
                    logger.error(err);
                    return -1;
                });
            }).catch(function(err) {
                logger.error(err);
                return -1;
            });
    }

    /**
     * Returns all clinics in the table.  Used for Message maintenance.
     */
    app.post('/api/getAllClinics', function(req,res){
       models.ClinicProfile.findAll({
           attributes: ['id', 'clinicId']
       }).then(function(data) {
           res.send(data);
       }).catch(function(err){
           res.status(500).send(err);
       })
    });

    /**
     * Return the count of NDC (POC) data that meet the filter criteria
     * (for pagination)
     *
     * Params:  (Body) "where" object with attributes that match the NDC model
     */
    app.post('/api/getNdcDataTotal/', function(req, res) {
        var whereArgs = buildNdcWhereArgs(req);

        models.NdcData.count({
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset
            }
        ).then(function(count) {
            res.send(String(count));
        })
    });

    /**
     * Return all NDC (POC) data that meet the filter criteria
     *
     * Params:  (Body) "where" object with attributes that match the NDC model
     */
    app.post('/api/getNdcData/', function(req, res) {
        var whereArgs = buildNdcWhereArgs(req);


        models.NdcData.findAll(
            {
                where: whereArgs,
                limit: req.body.limit,
                offset: req.body.offset,
                order: req.body.order
            }

        ).then(function (ndcData) {
            res.send(ndcData);
        });
    });

    app.post('/api/getAsnData', function(req, res) {
        var helper = new AsnHelper(configJson);
        helper.getAsnDataPromise(req).then(function (response) {
            res.header("X-Total-Count", response.headers['x-total-count']);
            res.status(200).send(response.body);
        }).catch(function (err) {
            res.status(500).send(err);
        })
    });

    app.post('/api/v1/getAsnFile', function(req, res) {
        var helper = new AsnHelper(configJson);
        helper.getAsnDataPromise(req).then(function (response) {
            var fields = ['asnId', 'asnDate', 'clinicId', 'poNumber', 'lineNumber', 'sapMaterialId', 'vendorItemId', 'description', 'shipVia', 'shipQty', 'receivedQty'];
            json2csv({data: response.body, fields: fields, del: ','}, function (err, csv) {
                if (err) logger.error(err);

                res.set({
                    'X-Total-Count': response.headers['x-total-count'],
                    'Content-Disposition': 'attachment; filename=ASN_data.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        });
    });

    //apis for downloading csv files
    function filterRow(row) {
        if (row.expiration != null) {
            var date = "\"" + moment(row.expiration).format("MM/YY") + "\"";
            row.expiration = date;
        }
        if (row.cnt_ind != null) {
            if (row.cnt_ind == 1) {
                row.cnt_ind = "X";
            } else {
                row.cnt_ind = "";
            }
        }

        //strip non-numerics from clinic ID (deprecated)
        //if (row.clinicId != null) {
        //    var clinicId = row.clinicId.replace(/[^0-9]/g, '');
        //    row.clinicId = clinicId;
        //}
    }

    function filterClinicRow(row, vendors) {
        for (var field in row) {
            if (String(field) == "catalogs") {
                //strip the default vendors from the field
                for (var v=0; v < vendors.length; v++) {
                    if (String(row[field]).indexOf(",") >= 0) {
                        row[field] = String(row[field]).replace("," + vendors[v].vendorId, '');
                        row[field] = String(row[field]).replace(vendors[v].vendorId + ",", '');
                    } else {
                        row[field] = String(row[field]).replace(vendors[v].vendorId, '');
                    }
                }
            } else {
                if (String(row[field])==="false") row[field] = 'n';
                if (String(row[field])==="true") row[field] = 'y';
            }

        }
    }

    function filterClinicRowSequelize(row, vendors, fields) {
        var data = {};
        _.forEach( fields, function(field) {
            if (field == "catalogs" && row.Catalogs) {
                data[field] = _.map(row.Catalogs, 'vendorId').join(",");
                //strip the default vendors from the field
                for (var v=0; v < vendors.length; v++) {
                    if (String(data[field]).indexOf(",") >= 0) {
                        data[field] = String(data[field]).replace("," + vendors[v].vendorId, '');
                        data[field] = String(data[field]).replace(vendors[v].vendorId + ",", '');
                    } else {
                        data[field] = String(data[field]).replace(vendors[v].vendorId, '');
                    }
                }
            } else {
                data[field] = row[field];
                if (String(row[field])==="false") data[field] = 'n';
                if (String(row[field])==="true") data[field] = 'y';
            }
        });

        return data;
    }

    //Data cleanup for the Master Item Export File
    function filterItemRow(row, fields) {
        var data = {};
        _.forEach( fields, function(field) {
            data[field] = (row[field]) ? row[field] : "";
        });

        if (row.cntInd != null) {
            if (data.cntInd == 1) {
                data.cntInd = "X";
            } else {
                data.cntInd = "";
            }
        }
        data['onhand_inventory_ind'] = data.showInventoryInd;

        if (row.Catalogs) {
            data['medVendor'] = row.catalogAsString;
        }
        else {
            data['medVendor'] = "";
        }

        if (data.discontinued === null || !data.discontinued) {
            data.discontinued = "";
        } else if (row.discontinued && row.discontinued.__proto__ && row.discontinued.__proto__.isValid ) {
            if (!row.discontinued.isValid()) {
                data.discontinued = "";
            }
        }

        if (data.alternates == null) {
            data.alternates = "";
        }

        if (data.ndc == null) {
            data.ndc = "";
        }

        if (data.unitPrice == null) {
            data.unitPrice = "";
        }

        if (data.vendorId == null) {
            data.vendorId = "";
        }

        return data;
    }

    app.post('/api/getCountFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;
            request.query('select time_stamp, clinic_id, sap_material, vendor_material, material_description, count, count_type, cat_type from tbl_count_data');

            var data = [];
            request.on('row', function(row) {
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("error: " + err);
            });
            request.on('done', function(returnVal) {
                var fields = ['time_stamp', 'clinic_id', 'sap_material', 'vendor_material', 'material_description', 'count', 'count_type', 'cat_type'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);
                    res.set({
                        'Content-Disposition': 'attachment; filename=countData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getCountFile', function(req, res){
        var whereArgs = buildCountDataWhereArgs(req);

        models.CountData.findAll({ where: whereArgs, raw:true }).then(function(data){
            var fields = ['timestamp', 'clinicId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'countType', 'catType'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=countData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/getInventoryFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select time_stamp, clinic_id, user_id, sap_material, vendor_material, description, count, " +
                "count_type, lot, expiration from tbl_inventory where type = 'INV'");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['time_stamp', 'clinic_id', 'user_id', 'sap_material', 'vendor_material', 'description', 'count', 'count_type', 'lot', 'expiration'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=inventoryData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getInventoryFile', function(req, res){
        models.Inventory.findAll({ where:{ type: 'INV' }, raw:true }).then(function(data){

            _.forEach( data, function(row) {
                filterRow(row);
            });

            var fields = ['timestamp', 'clinicId', 'userId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'countType', 'lot', 'expiration'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=inventoryData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/v2/getMasterItemFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            var q = "select id, sap_material_id, material_description, cnt_ind, cnt_description, " +
                " ISNULL(STUFF((SELECT ',' + v.vendor_id FROM db_owner.tbl_vendor_list v, db_owner.tbl_item_catalog ic WHERE ic.item_id = mi.id and ic.vendor_id = v.id for xml path('')),1,1,''),'') as catalogs, " +
                "    vendor_material, conv, count_type, " +
                "    ISNULL(alternates,'') as alternates, onhand_inventory_ind, " +
                "    CASE WHEN discontinued IS NULL THEN '' ELSE CONVERT(VARCHAR,discontinued,21) END as discontinued" +
                " from db_owner.tbl_master_item mi";
            logger.debug(q);
            request.query(q);
            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['sap_material_id', 'material_description', 'cnt_ind', 'cnt_description', 'catalogs', 'vendor_material', 'conv', 'count_type', 'alternates', 'onhand_inventory_ind', 'discontinued'];
                json2csv({ data: data, fields: fields, del: '\t' }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=inventoryData.csv',
                        'Content-Type': 'text/tab-separated-values'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v3/getMasterItemFile', function(req, res){
        models.MasterItem.findAll({ include: [{ model: models.CatalogList, as: 'Catalogs' }],
            paranoid:false }).then(function(data){
            var cleanData = [];
            var fields = ['sapMaterial', 'description', 'cntInd', 'cntDescription', 'medVendor', 'vendorMaterial', 'conv', 'countType', 'alternates', 'unitPrice', 'vendorId', 'ndc', 'onhand_inventory_ind', 'discontinued'];

            _.forEach( data, function(row) {
                cleanData.push(filterItemRow(row, fields));
            });

            json2csv({ data: cleanData, fields: fields, del: '\t' }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=itemData.tsv',
                    'Content-Type': 'text/tsv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/v1/getMasterItemMessageFile', function(req, res){
        var alternates = [];
        alternates.push({alternates: {
            $not: null
        }});
        alternates.push({alternates: {
            $not: ""
        }});
        var messages = [];
        messages.push({
            message: {
                $not: null
            }
        });
        messages.push({
            message: {
                $not: ""
            }
        });

        models.MasterItem.findAll({
            where: {
                $or: [
                    {
                        $and: alternates
                    },
                    {
                        $and: messages
                    }
                ]
            },
            include: [{ model: models.CatalogList, as: 'Catalogs' }],
            paranoid:false
        }).then(function(data){
            var cleanData = [];
            var fields = ['sapMaterial', 'description', 'vendorMaterial', 'alternates', 'message', 'discontinued'];

            _.forEach( data, function(row) {
                cleanData.push(filterItemRow(row, fields));
            });

            json2csv({ data: cleanData, fields: fields, del: '\t' }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=itemMessageData.tsv',
                    'Content-Type': 'text/tsv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/getMasterItemFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select sap_material_id, material_description, cnt_ind, cnt_description, med_vendor, vendor_material, conv, " +
                "count_type, CASE WHEN alternates IS NULL THEN '' ELSE alternates END as alternates, onhand_inventory_ind, CASE WHEN discontinued IS NULL THEN '' ELSE CONVERT(VARCHAR,discontinued,21) END as discontinued from tbl_master_item mi");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['sap_material_id', 'material_description', 'cnt_ind', 'cnt_description', 'med_vendor', 'vendor_material', 'conv', 'count_type', 'alternates', 'onhand_inventory_ind', 'discontinued'];
                json2csv({ data: data, fields: fields, del: '\t' }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=inventoryData.csv',
                        'Content-Type': 'text/tab-separated-values'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getClinicProfileFile', function(req, res) {
        models.VendorList.findAll({ where: {isDefault: true} }).then(function(vendors) {
            var conn = new mssql.Connection(config, function(err) {
                conn.connect();
                var request = new mssql.Request(conn);
                request.stream = true;
                var q = "SELECT clinic_id, ISNULL(STUFF((SELECT ',' + v.vendor_id FROM db_owner.tbl_vendor_list v, db_owner.tbl_clinic_catalog cc WHERE cc.clinic_id = cp.id and cc.vendor_id = v.id for xml path('')),1,1,''),'') as catalogs, " +
                    " mircera_visible, mircera_io, areanesp_visible as aranesp_visible, areanesp_io as aranesp_io, epogen_visible, epogen_io, count_in_each " +
                    " from tbl_clinic_profile cp order by clinic_id asc";
                logger.debug(q);
                request.query(q);

                var data = [];
                var idx = 0;
                var defVendorField = 'default_catalogs';
                var instructionsField = "For best results, open a blank worksheet and import this file";
                request.on('row', function(row) {

                    filterClinicRow(row, vendors);
                    if (idx < vendors.length) {
                        row[defVendorField] = vendors[idx].vendorId;
                    } else {
                        row[defVendorField] = "";
                    }
                    idx++;
                    data.push(row);
                });
                request.on('error', function(err) {
                    logger.error("error: " + err);
                });
                request.on('done', function(returnVal) {

                    var fields = ['clinic_id', 'catalogs', 'mircera_visible', 'mircera_io', 'aranesp_visible', 'aranesp_io', 'epogen_visible', 'epogen_io', 'count_in_each', defVendorField, '', instructionsField];
                    json2csv({ data: data, fields: fields, del: '\t' }, function(err, csv) {
                        if (err) logger.error(err);

                        res.set({
                            'Content-Disposition': 'attachment; filename=CHIPS_Locator_Data.tsv',
                            'Content-Type': 'text/tab-separated-values'
                        });
                        res.end(csv);
                    });
                });
            });
        });
    });

    app.post('/api/v3/getClinicProfileFile', function(req, res) {
        models.CatalogList.findAll({ where: {isDefault: true} }).then(function(vendors) {
            models.ClinicProfile.findAll({
                include: [{ model: models.CatalogList, as: 'Catalogs' }],
                order: [
                    ['clinicId', 'ASC']]}).then(function(clinics) {

                var defVendorField = 'default_catalogs';
                var instructionsField = "For best results, open a blank worksheet and import this file";
                var cleanData = [];
                var fields = ['clinicId', 'catalogs', 'mirceraVisible', 'mirceraIo', 'areanespVisible', 'areanespIo', 'epogenVisible', 'epogenIo', 'countInEach'];
                for (var idx = 0; idx < clinics.length; idx++) {
                    var row = clinics[idx];
                    var newRow = filterClinicRowSequelize(row, vendors, fields);
                    if (idx < vendors.length) {
                        newRow[defVendorField] = vendors[idx].vendorId;
                    } else {
                        newRow[defVendorField] = "";
                    }
                    cleanData.push(newRow);
                }
                fields = _.concat(fields, defVendorField, '', instructionsField);
                json2csv({ data: cleanData, fields: fields, del: '\t' }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=CHIPS_Locator_Data.tsv',
                        'Content-Type': 'text/tab-separated-values'
                    });
                    res.end(csv);
                });

            }).catch(function(cerr){
                logger.error(cerr);
            });
        }).catch(function(verr){
            logger.error(verr);
        });
    });

    app.post('/api/getClinicProfileFile', function(req, res) {
        models.VendorList.findAll({ where: {isDefault: true} }).then(function(vendors) {
            var conn = new mssql.Connection(config, function(err) {
                conn.connect();
                var request = new mssql.Request(conn);
                request.stream = true;

                request.query("SELECT clinic_id, vendor_list, mircera_visible, mircera_io, areanesp_visible as aranesp_visible, areanesp_io as aranesp_io, epogen_visible, epogen_io, count_in_each " +
                    " from tbl_clinic_profile  order by clinic_id asc");

                var data = [];
                var idx = 0;
                var defVendorField = 'default_vendors';
                var instructionsField = "For best results, open a blank worksheet and import this file";
                request.on('row', function(row) {

                    filterClinicRow(row, vendors);
                    if (idx < vendors.length) {
                        row[defVendorField] = vendors[idx].vendorId;
                    } else {
                        row[defVendorField] = "";
                    }
                    idx++;
                    data.push(row);
                });
                request.on('error', function(err) {
                    logger.error("error: " + err);
                });
                request.on('done', function(returnVal) {

                    var fields = ['clinic_id', 'vendor_list', 'mircera_visible', 'mircera_io', 'aranesp_visible', 'aranesp_io', 'epogen_visible', 'epogen_io', 'count_in_each', defVendorField, '', instructionsField];
                    json2csv({ data: data, fields: fields, del: '\t' }, function(err, csv) {
                        if (err) logger.error(err);

                        res.set({
                            'Content-Disposition': 'attachment; filename=CHIPS_Locator_Data.tsv',
                            'Content-Type': 'text/tab-separated-values'
                        });
                        res.end(csv);
                    });
                });
            });
        });
    });

    app.post('/api/getPoFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select time_stamp, clinic_id, user_id, sap_material, vendor_material, description, count, " +
                "count_type, lot, expiration from tbl_inventory where type = 'PO'");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("request error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['time_stamp', 'clinic_id', 'user_id', 'sap_material', 'vendor_material', 'description', 'count', 'count_type', 'lot', 'expiration'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=inventoryData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getPoFile', function(req, res){
        models.Inventory.findAll({ where:{ type: 'PO' }, raw:true }).then(function(data){

            _.forEach( data, function(row) {
                filterRow(row);
            });

            var fields = ['timestamp', 'clinicId', 'userId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'countType', 'lot', 'expiration'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=poData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/getReceiptFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select time_stamp, clinic_id, user_id, sap_material, vendor_material, description, count, " +
                "po_prefix, po_number, lot, expiration from tbl_inventory where type = 'RECEIPT'");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("request error: " + err);
            });
            request.on('done', function(returnVal) {
                var fields = ['time_stamp', 'clinic_id', 'user_id', 'sap_material', 'vendor_material', 'description', 'count', 'po_prefix', 'po_number', 'lot', 'expiration'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=receiptData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getReceiptFile', function(req, res){
        var whereArgs = buildInvWhereArgs(req);
        whereArgs.push({type: 'RECEIPT'});
        models.Inventory.findAll({ where: whereArgs, raw:true }).then(function(data){

            _.forEach( data, function(row) {
                filterRow(row);
            });

            var fields = ['timestamp', 'clinicId', 'userId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'poPrefix', 'poNumber', 'lot', 'expiration'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=receiptData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/getConsumptionFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select time_stamp, clinic_id, user_id, sap_material, vendor_material, description, count, " +
                "lot, expiration from tbl_inventory where type = 'CONSUMERUSAGE'");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("request error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['time_stamp', 'clinic_id', 'user_id', 'sap_material', 'vendor_material', 'description', 'count', 'lot', 'expiration'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=consumptionData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getConsumptionFile', function(req, res){
        models.Inventory.findAll({ where:{ type: 'CONSUMERUSAGE' }, raw:true }).then(function(data){

            _.forEach( data, function(row) {
                filterRow(row);
            });

            var fields = ['timestamp', 'clinicId', 'userId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'countType', 'lot', 'expiration'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=consumptionData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });



};

/**
 * Require authentication for ALL methods except GET on favicon.
 * @param req Request object
 * @param res Response object
 * @param next chained callback
 */
var auth = function(req, res, next) {
    if (req.url == "assets/images/favicon.png" || req.url == "/favicon.ico") {
        next();
    }
    else if (!req.isAuthenticated())
        res.status(401).send('Please log in');
    else next();
};



